Imports DES.PopulationGrowth.Events
Imports MathNet.Numerics.Distributions

Namespace DES.PopulationGrowth.Objects
	Public MustInherit Class Individual
		Public Property Age As Integer

		Public Property RelationAge As Integer

		Public Property LifeTime As Integer

		Public Property TimeChildren As Double

		Public Property Couple As Individual

		Protected Sub New(ByVal age As Integer)
			Me.Age = age
		End Sub

        Public Sub FindCouple(ByVal population As IEnumerable(Of Individual), ByVal currentTime As Integer, ByVal distributions As Dictionary(Of [Event], IDistribution))
            For Each candidate In population
                If SuitablePartner(candidate) AndAlso candidate.SuitableRelation AndAlso (CType(distributions([Event].BirthEngageDisengage), ContinuousUniform)).Sample() <= 0.5 Then
                    ' Relate them
                    candidate.Couple = Me
                    Couple = candidate
                    ' Set time for having child
                    Dim childTime = (CType(distributions([Event].TimeChildren), Exponential)).Sample * 100
                    ' They can have children on the simulated year: 'currentTime + childTime'.
                    candidate.TimeChildren = currentTime + childTime
                    TimeChildren = currentTime + childTime
                    Exit For
                End If
            Next candidate
        End Sub

        Public Function SuitableRelation() As Boolean
			Return Age >= RelationAge AndAlso Couple Is Nothing
		End Function

		Public Function SuitablePartner(ByVal individual As Individual) As Boolean
			Return ((TypeOf individual Is Male AndAlso TypeOf Me Is Female) OrElse (TypeOf individual Is Female AndAlso TypeOf Me Is Male)) AndAlso Math.Abs(individual.Age - Age) <= 5
		End Function

		Public ReadOnly Property Engaged As Boolean
			Get
				Return Couple IsNot Nothing
			End Get
		End Property

		Public Sub Disengage()
			Couple.Couple = Nothing
			Couple = Nothing
			TimeChildren = 0
		End Sub

        Public Function EndRelation(ByVal distributions As Dictionary(Of [Event], IDistribution)) As Boolean
            Dim sample = (CType(distributions([Event].BirthEngageDisengage), ContinuousUniform)).Sample

            If Age >= 14 AndAlso Age <= 20 AndAlso sample <= 0.7 Then
                Return True
            End If
            If Age >= 21 AndAlso Age <= 28 AndAlso sample <= 0.5 Then
                Return True
            End If
            If Age >= 29 AndAlso sample <= 0.2 Then
                Return True
            End If

            Return False
        End Function

        Public Overrides Function ToString() As String
			Return String.Format("Age: {0} Lifetime {1}", Age, LifeTime)
		End Function
	End Class
End Namespace
