Imports DES.PopulationGrowth.Events
Imports DES.PopulationGrowth.Objects
Imports MathNet.Numerics.Distributions
Imports MathNet.Numerics.Random

'using Puzzle.MicroHeap;

Namespace DES.PopulationGrowth
	Public Class Simulation
		Public Property Population As List(Of Individual)

		Public Property Time As Integer
		Private _currentTime As Integer
        Private ReadOnly _distributions As Dictionary(Of [Event], IDistribution)

        Public Sub New(ByVal population As IEnumerable(Of Individual), ByVal time As Integer)
			Me.Population = New List(Of Individual)(population)
			Me.Time = time
            _distributions = New Dictionary(Of [Event], IDistribution) From {{[Event].CapableEngaging, New Poisson(18)}, {[Event].BirthEngageDisengage, New ContinuousUniform}, {[Event].GetPregnant, New Normal(28, 8)}, {[Event].ChildrenCount, New Normal(2, 6)}, {[Event].TimeChildren, New Exponential(8)}, {[Event].Die, New Poisson(70)}}

            For Each individual In Me.Population
                ' LifeTime
                individual.LifeTime = (CType(_distributions([Event].Die), Poisson)).Sample
                ' Ready to start having relations
                individual.RelationAge = (CType(_distributions([Event].CapableEngaging), Poisson)).Sample
                ' Pregnancy Age (only women)
                If TypeOf individual Is Female Then
                    TryCast(individual, Female).PregnancyAge = (CType(_distributions([Event].GetPregnant), Normal)).Sample
                    TryCast(individual, Female).ChildrenCount = (CType(_distributions([Event].ChildrenCount), Normal)).Sample
                End If
			Next individual
		End Sub

		Public Sub Execute()
			Do While _currentTime < Time
				' Check what happens to every individual this year
				Dim i = 0
				Do While i < Population.Count
					Dim individual = Population(i)

					' Event -> Birth
					If TypeOf individual Is Female AndAlso (TryCast(individual, Female)).IsPregnant Then
						Population.Add((TryCast(individual, Female)).GiveBirth(_distributions, _currentTime))
					End If

					' Event -> Check whether someone starts a relation this year
					If individual.SuitableRelation Then
						individual.FindCouple(Population, _currentTime, _distributions)
					End If

					' Events where having an engaged individual represents a prerequisite
					If individual.Engaged Then
						' Event -> Check whether some relation ends this year
						If individual.EndRelation(_distributions) Then
							individual.Disengage()
						End If

						' Event -> Check whether some couple can have a child now
						If TypeOf individual Is Female AndAlso (TryCast(individual, Female)).SuitablePregnancy(_currentTime) Then
							TryCast(individual, Female).IsPregnant = True
						End If
					End If

					' Event -> Check whether someone dies this year
					If individual.Age.Equals(individual.LifeTime) Then
						' Case: Individual in relation (break relation)
						If individual.Engaged Then
							individual.Disengage()
						End If
						Population.RemoveAt(i)
					End If

					individual.Age += 1
					_currentTime += 1
					i += 1
				Loop
			Loop
		End Sub
	End Class
End Namespace
